# from Probability and Statistical Inference: Hogg & Tanis, Example 6.3.2
# submitted by Ashvin Goel
prog='dbrvstatdiff'
args='mean2 stddev2 n2 mean1 stddev1 n1'
cmp='diff -c -b '
altcmp='dbfilediff --quiet -E -N test_diff --exit '
altcmp_needs_input_flags=true
